@@ -117,6 +117,9 @@ class AbstractBackend:
     def __repr__(self):
         return "<einops backend for {}>".format(self.framework_name)
 
+    def einsum(self, pattern, *x):
+        raise NotImplementedError("backend does not support einsum")
+
 
 class UnknownSize:
     """ pseudo-symbol for symbolic frameworks which do not provide symbols for shape elements """
@@ -163,11 +166,14 @@ class NumpyBackend(AbstractBackend):
         return self.np.tile(x, repeats)
 
     def is_float_type(self, x):
-        return x.dtype in ('float16', 'float32', 'float64', 'float128')
+        return x.dtype in ('float16', 'float32', 'float64', 'float128', 'bfloat16')
 
     def add_axis(self, x, new_position):
         return self.np.expand_dims(x, new_position)
 
+    def einsum(self, pattern, *x):
+        return self.np.einsum(pattern, *x)
+
 
 class JaxBackend(NumpyBackend):
     framework_name = 'jax'
@@ -333,6 +339,13 @@ class TorchBackend(AbstractBackend):
     def stack_on_zeroth_dimension(self, tensors: list):
         return self.torch.stack(tensors)
 
+    def add_axes(self, x, n_axes, pos2len):
+        repeats = [-1] * n_axes
+        for axis_position, axis_length in pos2len.items():
+            x = self.add_axis(x, axis_position)
+            repeats[axis_position] = axis_length
+        return x.expand(repeats)
+
     def tile(self, x, repeats):
         return x.repeat(repeats)
 
@@ -340,12 +353,15 @@ class TorchBackend(AbstractBackend):
         return self.torch.unsqueeze(x, new_position)
 
     def is_float_type(self, x):
-        return x.dtype in [self.torch.float16, self.torch.float32, self.torch.float64]
+        return x.dtype in [self.torch.float16, self.torch.float32, self.torch.float64, self.torch.bfloat16]
 
     def layers(self):
         from .layers import torch
         return torch
 
+    def einsum(self, pattern, *x):
+        return self.torch.einsum(pattern, *x)
+
 
 class CupyBackend(AbstractBackend):
     framework_name = 'cupy'
@@ -376,7 +392,10 @@ class CupyBackend(AbstractBackend):
         return self.cupy.expand_dims(x, new_position)
 
     def is_float_type(self, x):
-        return x.dtype in ('float16', 'float32', 'float64', 'float128')
+        return x.dtype in ('float16', 'float32', 'float64', 'float128', 'bfloat16')
+
+    def einsum(self, pattern, *x):
+        return self.cupy.einsum(pattern, *x)
 
 
 class ChainerBackend(AbstractBackend):
@@ -415,12 +434,15 @@ class ChainerBackend(AbstractBackend):
         return self.chainer.functions.expand_dims(x, new_position)
 
     def is_float_type(self, x):
-        return x.dtype in ('float16', 'float32', 'float64', 'float128')
+        return x.dtype in ('float16', 'float32', 'float64', 'float128', 'bfloat16')
 
     def layers(self):
         from .layers import chainer
         return chainer
 
+    def einsum(self, pattern, *x):
+        return self.chainer.functions.einsum(pattern, *x)
+
 
 class HashableTuple:
     """Overcomes non-hashability of symbolic elements"""
@@ -494,12 +516,15 @@ class TensorflowBackend(AbstractBackend):
         return self.tf.expand_dims(x, new_position)
 
     def is_float_type(self, x):
-        return x.dtype in ('float16', 'float32', 'float64', 'float128')
+        return x.dtype in ('float16', 'float32', 'float64', 'float128', 'bfloat16')
 
     def layers(self):
         from .layers import tensorflow
         return tensorflow
 
+    def einsum(self, pattern, *x):
+        return self.tf.einsum(pattern, *x)
+
 
 class KerasBackend(AbstractBackend):
     framework_name = 'tensorflow.keras'
@@ -517,9 +542,8 @@ class KerasBackend(AbstractBackend):
         return self.keras.Input(batch_shape=shape)
 
     def eval_symbol(self, symbol, input_dict):
-        (variable, value), = input_dict
-        model = self.keras.models.Model(variable, symbol)
-        return model.predict_on_batch(value)
+        model = self.keras.models.Model([var for (var, _) in input_dict], symbol)
+        return model.predict_on_batch([val for (_, val) in input_dict])
 
     def arange(self, start, stop):
         return self.K.arange(start, stop)
@@ -552,3 +576,69 @@ class KerasBackend(AbstractBackend):
     def layers(self):
         from .layers import keras
         return keras
+
+
+class OneFlowBackend(AbstractBackend):
+    framework_name = "oneflow"
+
+    def __init__(self):
+        import oneflow as flow
+        self.flow = flow
+        print("using oneflow")
+
+    def is_appropriate_type(self, tensor):
+        return isinstance(tensor, self.flow.Tensor)
+
+    def from_numpy(self, x):
+        variable = self.flow.from_numpy(x)
+        if self.is_float_type(variable):
+            # attach grad only to floating types
+            variable.requires_grad = True
+        return variable
+
+    def to_numpy(self, x):
+        return x.detach().cpu().numpy()
+
+    def arange(self, start, stop):
+        return self.flow.arange(start, stop, dtype=self.flow.int64)
+
+    def reduce(self, x, operation, reduced_axes):
+        for axis in sorted(reduced_axes, reverse=True):
+            if operation == 'min':
+                x, _ = x.min(dim=axis)
+            elif operation == 'max':
+                x, _ = x.max(dim=axis)
+            elif operation in ['sum', 'mean', 'prod']:
+                x = getattr(x, operation)(dim=axis)
+            else:
+                raise NotImplementedError('Unknown reduction ', operation)
+        return x
+
+    def transpose(self, x, axes):
+        return x.permute(axes)
+
+    def stack_on_zeroth_dimension(self, tensors: list):
+        return self.flow.stack(tensors)
+
+    def add_axes(self, x, n_axes, pos2len):
+        repeats = [-1] * n_axes
+        for axis_position, axis_length in pos2len.items():
+            x = self.add_axis(x, axis_position)
+            repeats[axis_position] = axis_length
+        return x.expand(*repeats)
+
+    def tile(self, x, repeats):
+        return x.repeat(repeats)
+
+    def add_axis(self, x, new_position):
+        return self.flow.unsqueeze(x, new_position)
+
+    def is_float_type(self, x):
+        return x.dtype in [self.flow.float16, self.flow.float32, self.flow.float64]
+
+    def layers(self):
+        from .layers import oneflow
+        return oneflow
+
+    def einsum(self, pattern, *x):
+        return self.flow.einsum(pattern, *x)
\ No newline at end of file
